class IdealBotCounter extends Mutator config;

// These values are used if the map doesn't contain a parseable range.
var() config int DefaultIdealMinPlayers;
var() config int DefaultIdealMaxPlayers;

// In case the map gives a stupidly high value, limit it to this.
var() config int MaxMinPlayers;

// This is the ideal number of players the server admin wants to have.
var() config int MinPlayers;

// These are computed once at the start of the game, and they don't
// change, unless a player reconfigures.
var int NumTeams;
var int IdealNumPlayers;

var bool bPreBeginPlayDone;

var int TeamSize[4]; // the number of human players on each team

function PreBeginPlay()
{
  Log("IBC::PreBeginPlay at "$Level.Game.GameReplicationInfo.ElapsedTime);

  // We only want to do things the first time this is called.  (Or do
  // we?  Maybe only the second time is useful to us.)
  if (bPreBeginPlayDone)
    return;
  bPreBeginPlayDone = True;

  if (Level.Game.bTeamGame) {
    // How many teams are there?
    NumTeams = 2; // Always for UT2003+?
  } else {
    NumTeams = 0;
  }

  SetIdealPlayerCount(Level.IdealPlayerCountMin, Level.IdealPlayerCountMax);
}


function GetRange(out int MinValue,
		  out int MaxValue,
		  coerce string Text,
		  int DefaultMin,
		  int DefaultMax)
{
  local string ClampRange, Word;
  local int DashPos;

  ClampRange = Text;
  DashPos = InStr(ClampRange, "-");
  if (DashPos != -1) {
    Word = Left(ClampRange, DashPos);
    MinValue = int(Word);
    Word = Mid(ClampRange, DashPos + 1);
    MaxValue = int(Word);
  } else {
    MinValue = DefaultMin;
    MaxValue = DefaultMax;
  }
}

function ExtractIdealPlayerCount(coerce string IPC)
{
  local int IdealMinPlayers;
  local int IdealMaxPlayers;

  // Get the range for clamping the number of players.
  GetRange(IdealMinPlayers, IdealMaxPlayers, IPC,
	   DefaultIdealMinPlayers, DefaultIdealMaxPlayers);
  SetIdealPlayerCount(IdealMinPlayers, IdealMaxPlayers);
}

function BroadcastMessage(coerce string Message, optional name Type)
{
  Log("IBC:"@Message);
  //Level.Game.BroadcastHandler.BroadcastText(None, None, Message, Type);
}

function SetIdealPlayerCount(int IdealMinPlayers,
			     int IdealMaxPlayers)
{
  // Don't let the map give stupid values.
  IdealMinPlayers = Min(IdealMinPlayers, MaxMinPlayers);

  // Clamp our suggested minimum number of players.
  IdealNumPlayers = Clamp(MinPlayers, IdealMinPlayers, IdealMaxPlayers);

  // Report the sanity-checked range pulled from the map.
  BroadcastMessage("Ideal player range="$IdealMinPlayers$
		   "-"$IdealNumPlayers$"-"$IdealMaxPlayers);
}





function Mutate(string MutateString, PlayerController Sender)
{
  Super.Mutate(MutateString, Sender);

  if (Left(Caps(MutateString), 5) == "RANGE") {
    ExtractIdealPlayerCount(Mid(MutateString, 5));
    if (NumTeams == 0)
      UnrealMPGameInfo(Level.Game).MinPlayers = IdealNumPlayers;
  }
}

function PostBeginPlay()
{
  if (NumTeams == 0) {
    // Set the minimum number of players now, and be done with it.
    UnrealMPGameInfo(Level.Game).MinPlayers = IdealNumPlayers;
  } else {
    // Establish periodic computation of MinPlayers.
    SetTimer(5.0, True);
  }
}

/* How many players do we want to balance the number of players in all
   teams with bots?  Divide the result by NumTeams to get the uniform
   team size. */
function SetOptimumSize()
{
  local int TeamIndex;
  local int IdealSize;

  // Find the biggest team.
  IdealSize = 0;
  for (TeamIndex = 0; TeamIndex < NumTeams; TeamIndex++)
    if (TeamSize[TeamIndex] > IdealSize)
      IdealSize = TeamSize[TeamIndex];
  //BroadcastMessage("Largest team has"@IdealSize);

  // We should be able to fill the smaller teams to make them as big
  // as the largest.
  IdealSize = NumTeams * IdealSize;
  //BroadcastMessage("Total size would be"@IdealSize);

  // We should bump this up according to the map's own recommended
  // size.
  if (IdealSize < IdealNumPlayers)
    IdealSize = IdealNumPlayers;
  //BroadcastMessage("Reaching min"@IdealNumPlayers@"would be"@IdealSize);

  // We should round this number up to a multiple of the number of
  // teams.
  IdealSize = (IdealSize + NumTeams - 1) / NumTeams * NumTeams;
  //BroadcastMessage("Even for all teams"@IdealSize);

  // Let's not go overboard.
  if (IdealSize > 32)
    IdealSize = 32 / NumTeams * NumTeams;

  //BroadcastMessage("Absolute limit"@IdealSize);

  UnrealMPGameInfo(Level.Game).MinPlayers = IdealSize;
}



function GetTeamSizes()
{
  local int TeamIndex;
  local PlayerController P;
  local TeamInfo Info;

  for (TeamIndex = 0; TeamIndex < NumTeams; TeamIndex++)
    TeamSize[TeamIndex] = 0;

  foreach AllActors (class'PlayerController', P) {
    Info = P.PlayerReplicationInfo.Team;
    if (Info != None) {
      TeamIndex = Info.TeamIndex;
      TeamSize[TeamIndex]++;
    }
  }
}


/* Periodically check that we don't have too many or too few bots. */
event Timer()
{
  GetTeamSizes();
  SetOptimumSize();
}


/**************** ****************/

defaultproperties
{
  DefaultIdealMinPlayers=4
  DefaultIdealMaxPlayers=8
  MaxMinPlayers=32
  MinPlayers=8
  FriendlyName="Ideal Bot Count"
  Description="Chooses the number of bots based on the map's recommendations."
  IconMaterialName="Ideal Bot Count xxx"
}
